﻿<#
    MIT License

    Copyright (c) Microsoft Corporation.

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE
#>

# Version 22.08.18.1501

#################################################################################
# Purpose:
# This script will allow you to test VSS functionality on Exchange server using DiskShadow.
# The script will automatically detect active and passive database copies running on the server.
# The general logic is:
# - start a PowerShell transcript
# - enable ExTRA tracing
# - enable VSS tracing
# - optionally: create the diskshadow config file with shadow expose enabled,
#               execute VSS backup using diskshadow,
#               delete the VSS snapshot post-backup
# - stop PowerShell transcript
#
#################################################################################
[Diagnostics.CodeAnalysis.SuppressMessageAttribute('PSAvoidUsingEmptyCatchBlock', '', Justification = 'Allowing empty catch blocks for now as we need to be able to handle the exceptions.')]
[CmdletBinding()]
param(
)



function Confirm-ProxyServer {
    [CmdletBinding()]
    [OutputType([bool])]
    param (
        [Parameter(Mandatory = $true)]
        [string]
        $TargetUri
    )

    try {
        $proxyObject = ([System.Net.WebRequest]::GetSystemWebproxy()).GetProxy($TargetUri)
        if ($TargetUri -ne $proxyObject.OriginalString) {
            return $true
        } else {
            return $false
        }
    } catch {
        return $false
    }
}

function Invoke-WebRequestWithProxyDetection {
    [CmdletBinding()]
    param (
        [Parameter(Mandatory = $true)]
        [string]
        $Uri,

        [Parameter(Mandatory = $false)]
        [switch]
        $UseBasicParsing,

        [Parameter(Mandatory = $false)]
        [string]
        $OutFile
    )

    [Net.ServicePointManager]::SecurityProtocol = [Net.SecurityProtocolType]::Tls12
    if (Confirm-ProxyServer -TargetUri $Uri) {
        $webClient = New-Object System.Net.WebClient
        $webClient.Headers.Add("User-Agent", "PowerShell")
        $webClient.Proxy.Credentials = [System.Net.CredentialCache]::DefaultNetworkCredentials
    }

    $params = @{
        Uri     = $Uri
        OutFile = $OutFile
    }

    if ($UseBasicParsing) {
        $params.UseBasicParsing = $true
    }

    Invoke-WebRequest @params
}

<#
    Determines if the script has an update available.
#>
function Get-ScriptUpdateAvailable {
    [CmdletBinding()]
    [OutputType([PSCustomObject])]
    param (
        [Parameter(Mandatory = $false)]
        [string]
        $VersionsUrl = "https://github.com/microsoft/CSS-Exchange/releases/latest/download/ScriptVersions.csv"
    )

    $BuildVersion = "22.08.18.1501"

    $scriptName = $script:MyInvocation.MyCommand.Name
    $scriptPath = [IO.Path]::GetDirectoryName($script:MyInvocation.MyCommand.Path)
    $scriptFullName = (Join-Path $scriptPath $scriptName)

    $result = [PSCustomObject]@{
        ScriptName     = $scriptName
        CurrentVersion = $BuildVersion
        LatestVersion  = ""
        UpdateFound    = $false
        Error          = $null
    }

    if ((Get-AuthenticodeSignature -FilePath $scriptFullName).Status -eq "NotSigned") {
        Write-Warning "This script appears to be an unsigned test build. Skipping version check."
    } else {
        try {
            $versionData = [Text.Encoding]::UTF8.GetString((Invoke-WebRequestWithProxyDetection $VersionsUrl -UseBasicParsing).Content) | ConvertFrom-Csv
            $latestVersion = ($versionData | Where-Object { $_.File -eq $scriptName }).Version
            $result.LatestVersion = $latestVersion
            if ($null -ne $latestVersion -and $latestVersion -ne $BuildVersion) {
                $result.UpdateFound = $true
            }

            Write-Verbose "Current version: $($result.CurrentVersion) Latest version: $($result.LatestVersion) Update found: $($result.UpdateFound)"
        } catch {
            Write-Verbose "Unable to check for updates: $($_.Exception)"
            $result.Error = $_
        }
    }

    return $result
}


function Invoke-CatchActionError {
    [CmdletBinding()]
    param(
        [scriptblock]$CatchActionFunction
    )

    if ($null -ne $CatchActionFunction) {
        & $CatchActionFunction
    }
}

function Invoke-CatchActionErrorLoop {
    [CmdletBinding()]
    param(
        [Parameter(Mandatory = $true, Position = 0)]
        [int]$CurrentErrors,
        [Parameter(Mandatory = $false, Position = 1)]
        [scriptblock]$CatchActionFunction
    )
    process {
        if ($null -ne $CatchActionFunction -and
            $Error.Count -ne $CurrentErrors) {
            $i = 0
            while ($i -lt ($Error.Count - $currentErrors)) {
                & $CatchActionFunction $Error[$i]
                $i++
            }
        }
    }
}

function Confirm-ExchangeShell {
    [CmdletBinding()]
    param(
        [Parameter(Mandatory = $false)]
        [ValidateNotNullOrEmpty()]
        [string]$Identity = $env:COMPUTERNAME,

        [Parameter(Mandatory = $false)]
        [bool]$LoadExchangeShell = $true,

        [Parameter(Mandatory = $false)]
        [scriptblock]$CatchActionFunction
    )

    begin {
        function Test-GetExchangeServerCmdletError {
            param(
                [Parameter(Mandatory = $true)]
                [object]$ThisError
            )

            if ($ThisError.FullyQualifiedErrorId -ne "CommandNotFoundException") {
                Write-Warning "Failed to find '$Identity' as an Exchange Server."
                return $true
            }
            return $false
        }
        $currentErrors = $Error.Count
        $passed = $false
        $edgeTransportKey = 'HKLM:\SOFTWARE\Microsoft\ExchangeServer\v15\EdgeTransportRole'
        $setupKey = 'HKLM:\SOFTWARE\Microsoft\ExchangeServer\v15\Setup'
        Write-Verbose "Calling: $($MyInvocation.MyCommand)"
        Write-Verbose "Passed: LoadExchangeShell: $LoadExchangeShell | Identity: $Identity"
        $params = @{
            Identity    = $Identity
            ErrorAction = "Stop"
        }

        $toolsServer = (Test-Path $setupKey) -and (!(Test-Path $edgeTransportKey)) -and `
        ($null -eq (Get-ItemProperty -Path $setupKey -Name "Services" -ErrorAction SilentlyContinue))
        $remoteShell = (!(Test-Path $setupKey))

        if ($toolsServer -or $remoteShell) {
            Write-Verbose "Exchange Server Check against local: $env:ComputerName"
            if ($env:ComputerName -eq $Identity -or
                $env:ComputerName -eq $Identity.Split(".")[0]) {
                Write-Verbose "Removing Identity from Get-ExchangeServer cmdlet due to passed value isn't an Exchange Server"
                $params.Remove("Identity")
            } else {
                Write-Verbose "Didn't remove Identity"
            }
        }

        Invoke-CatchActionErrorLoop $currentErrors $CatchActionFunction
    }
    process {
        try {
            $currentErrors = $Error.Count
            $isEMS = Get-ExchangeServer @params |
                Select-Object -First 1 |
                ForEach-Object { if ($_.GetType().Name -eq "ExchangeServer") { return $true } return $false }
            Write-Verbose "Exchange PowerShell Module already loaded."
            $passed = $true
            Invoke-CatchActionErrorLoop $currentErrors $CatchActionFunction
        } catch {
            Write-Verbose "Failed to run Get-ExchangeServer"
            Invoke-CatchActionError $CatchActionFunction
            if (Test-GetExchangeServerCmdletError $_) { return }
            if (-not ($LoadExchangeShell)) { return }

            #Test 32 bit process, as we can't see the registry if that is the case.
            if (-not ([System.Environment]::Is64BitProcess)) {
                Write-Warning "Open a 64 bit PowerShell process to continue"
                return
            }

            if (Test-Path "$setupKey") {
                $currentErrors = $Error.Count
                Write-Verbose "We are on Exchange 2013 or newer"

                try {
                    if (Test-Path $edgeTransportKey) {
                        Write-Verbose "We are on Exchange Edge Transport Server"
                        [xml]$PSSnapIns = Get-Content -Path "$env:ExchangeInstallPath\Bin\exshell.psc1" -ErrorAction Stop

                        foreach ($PSSnapIn in $PSSnapIns.PSConsoleFile.PSSnapIns.PSSnapIn) {
                            Write-Verbose ("Trying to add PSSnapIn: {0}" -f $PSSnapIn.Name)
                            Add-PSSnapin -Name $PSSnapIn.Name -ErrorAction Stop
                        }

                        Import-Module $env:ExchangeInstallPath\bin\Exchange.ps1 -ErrorAction Stop
                    } else {
                        Import-Module $env:ExchangeInstallPath\bin\RemoteExchange.ps1 -ErrorAction Stop
                        Connect-ExchangeServer -Auto -ClientApplication:ManagementShell
                    }

                    Write-Verbose "Imported Module. Trying Get-Exchange Server Again"
                    try {
                        $isEMS = Get-ExchangeServer @params |
                            Select-Object -First 1 |
                            ForEach-Object { if ($_.GetType().Name -eq "ExchangeServer") { return $true } return $false }
                        $passed = $true
                        Write-Verbose "Successfully loaded Exchange Management Shell"
                        Invoke-CatchActionErrorLoop $currentErrors $CatchActionFunction
                    } catch {
                        Write-Verbose "Failed to run Get-ExchangeServer again"
                        Invoke-CatchActionError $CatchActionFunction
                        if (Test-GetExchangeServerCmdletError $_) { return }
                    }
                } catch {
                    Write-Warning "Failed to Load Exchange PowerShell Module..."
                    Invoke-CatchActionError $CatchActionFunction
                }
            } else {
                Write-Verbose "Not on an Exchange or Tools server"
            }
        }
    }
    end {

        $currentErrors = $Error.Count
        $returnObject = [PSCustomObject]@{
            ShellLoaded = $passed
            Major       = ((Get-ItemProperty -Path $setupKey -Name "MsiProductMajor" -ErrorAction SilentlyContinue).MsiProductMajor)
            Minor       = ((Get-ItemProperty -Path $setupKey -Name "MsiProductMinor" -ErrorAction SilentlyContinue).MsiProductMinor)
            Build       = ((Get-ItemProperty -Path $setupKey -Name "MsiBuildMajor" -ErrorAction SilentlyContinue).MsiBuildMajor)
            Revision    = ((Get-ItemProperty -Path $setupKey -Name "MsiBuildMinor" -ErrorAction SilentlyContinue).MsiBuildMinor)
            EdgeServer  = $passed -and (Test-Path $setupKey) -and (Test-Path $edgeTransportKey)
            ToolsOnly   = $passed -and $toolsServer
            RemoteShell = $passed -and $remoteShell
            EMS         = $isEMS
        }

        Invoke-CatchActionErrorLoop $currentErrors $CatchActionFunction

        return $returnObject
    }
}

function Invoke-CreateDiskShadowFile {
    [Diagnostics.CodeAnalysis.SuppressMessageAttribute('PSAvoidUsingWMICmdlet', '', Justification = 'Required to get drives on old systems')]
    param()

    function Out-DHSFile {
        param ([string]$fileline)
        $fileline | Out-File -FilePath "$path\diskshadow.dsh" -Encoding ASCII -Append
    }

    #	creates the diskshadow.dsh file that will be written to below
    #	-------------------------------------------------------------
    $nl
    Get-Date
    Write-Host "Creating diskshadow config file..." -ForegroundColor Green $nl
    Write-Host "--------------------------------------------------------------------------------------------------------------"
    $nl
    New-Item -Path $path\diskshadow.dsh -type file -Force | Out-Null

    #	beginning lines of file
    #	-----------------------
    Out-DHSFile "set verbose on"
    Out-DHSFile "set context persistent"
    Out-DHSFile " "

    #	writers to exclude
    #	------------------
    Out-DHSFile "writer exclude {e8132975-6f93-4464-a53e-1050253ae220}"
    Out-DHSFile "writer exclude {2a40fd15-dfca-4aa8-a654-1f8c654603f6}"
    Out-DHSFile "writer exclude {35E81631-13E1-48DB-97FC-D5BC721BB18A}"
    Out-DHSFile "writer exclude {be000cbe-11fe-4426-9c58-531aa6355fc4}"
    Out-DHSFile "writer exclude {4969d978-be47-48b0-b100-f328f07ac1e0}"
    Out-DHSFile "writer exclude {a6ad56c2-b509-4e6c-bb19-49d8f43532f0}"
    Out-DHSFile "writer exclude {afbab4a2-367d-4d15-a586-71dbb18f8485}"
    Out-DHSFile "writer exclude {59b1f0cf-90ef-465f-9609-6ca8b2938366}"
    Out-DHSFile "writer exclude {542da469-d3e1-473c-9f4f-7847f01fc64f}"
    Out-DHSFile "writer exclude {4dc3bdd4-ab48-4d07-adb0-3bee2926fd7f}"
    Out-DHSFile "writer exclude {41e12264-35d8-479b-8e5c-9b23d1dad37e}"
    Out-DHSFile "writer exclude {12ce4370-5bb7-4C58-a76a-e5d5097e3674}"
    Out-DHSFile "writer exclude {cd3f2362-8bef-46c7-9181-d62844cdc062}"
    Out-DHSFile "writer exclude {dd846aaa-A1B6-42A8-AAF8-03DCB6114BFD}"
    Out-DHSFile "writer exclude {B2014C9E-8711-4C5C-A5A9-3CF384484757}"
    Out-DHSFile "writer exclude {BE9AC81E-3619-421F-920F-4C6FEA9E93AD}"
    Out-DHSFile "writer exclude {F08C1483-8407-4A26-8C26-6C267A629741}"
    Out-DHSFile "writer exclude {6F5B15B5-DA24-4D88-B737-63063E3A1F86}"
    Out-DHSFile "writer exclude {368753EC-572E-4FC7-B4B9-CCD9BDC624CB}"
    Out-DHSFile "writer exclude {5382579C-98DF-47A7-AC6C-98A6D7106E09}"
    Out-DHSFile "writer exclude {d61d61c8-d73a-4eee-8cdd-f6f9786b7124}"
    Out-DHSFile "writer exclude {75dfb225-e2e4-4d39-9ac9-ffaff65ddf06}"
    Out-DHSFile "writer exclude {0bada1de-01a9-4625-8278-69e735f39dd2}"
    Out-DHSFile "writer exclude {a65faa63-5ea8-4ebc-9dbd-a0c4db26912a}"
    Out-DHSFile " "

    #	add databases to exclude
    #	------------------------
    foreach ($db in $databases) {
        $dbg = ($db.guid)

        if (($db).guid -ne $dbGuid) {
            if (($db.IsMailboxDatabase) -eq "True") {
                $mountedOnServer = (Get-MailboxDatabase $db).server.name
            } else {
                $mountedOnServer = (Get-PublicFolderDatabase $db).server.name
            }
            if ($mountedOnServer -eq $serverName) {
                $script:activeNode = $true

                Out-DHSFile "writer exclude `"Microsoft Exchange Writer:\Microsoft Exchange Server\Microsoft Information Store\$serverName\$dbg`""
            }
            #if passive copy, add it with replica in the string
            else {
                $script:activeNode = $false
                Out-DHSFile "writer exclude `"Microsoft Exchange Replica Writer:\Microsoft Exchange Server\Microsoft Information Store\Replica\$serverName\$dbg`""
            }
        }
        #	add database to include
        #	-----------------------
        else {
            if (($db.IsMailboxDatabase) -eq "True") {
                $mountedOnServer = (Get-MailboxDatabase $db).server.name
            } else {
                $mountedOnServer = (Get-PublicFolderDatabase $db).server.name
            }
        }
    }
    Out-DHSFile " "
    Out-DHSFile "Begin backup"

    #	add the volumes for the included database
    #	-----------------------------------------
    #gets a list of mount points on local server
    $mpvolumes = Get-WmiObject -Query "select name, deviceid from win32_volume where drivetype=3 AND driveletter=NULL"
    $deviceIDs = @()

    #if selected database is a mailbox database, get mailbox paths
    if ((($databases[$dbtoBackup]).IsMailboxDatabase) -eq "True") {
        $getDB = (Get-MailboxDatabase $selDB)

        $dbMP = $false
        $logMP = $false

        #if no mountpoints ($mpvolumes) causes null-valued error, need to handle
        if ($null -ne $mpvolumes) {
            foreach ($mp in $mpvolumes) {
                $mpname = (($mp.name).substring(0, $mp.name.length - 1))
                #if following mount point path exists in database path use deviceID in diskshadow config file
                if ($getDB.EdbFilePath.pathname.ToString().ToLower().StartsWith($mpname.ToString().ToLower())) {
                    Write-Host " "
                    Write-Host "Mount point:  $($mp.name) in use for database path: "
                    #Write-host "Yes. I am a database in mountpoint"
                    "The selected database path is: " + $getDB.EdbFilePath.pathname
                    Write-Host "adding deviceID to file: "
                    $dbEdbVol = $mp.deviceid
                    Write-Host $dbEdbVol

                    #add device ID to array
                    $deviceID1 = $mp.DeviceID
                    $dbMP = $true
                }

                #if following mount point path exists in log path use deviceID in diskshadow config file
                if ($getDB.LogFolderPath.pathname.ToString().ToLower().contains($mpname.ToString().ToLower())) {
                    Write-Host " "
                    Write-Host "Mount point: $($mp.name) in use for log path: "
                    #Write-host "Yes. My logs are in a mountpoint"
                    "The log folder path of selected database is: " + $getDB.LogFolderPath.pathname
                    Write-Host "adding deviceID to file: "
                    $dbLogVol = $mp.deviceid
                    Write-Host $dbLogVol
                    $deviceID2 = $mp.DeviceID
                    $logMP = $true
                }
            }
            $deviceIDs = $deviceID1, $deviceID2
        }
    }

    #if not a mailbox database, assume its a public folder database, get public folder paths
    if ((($databases[$dbtoBackup]).IsPublicFolderDatabase) -eq "True") {
        $getDB = (Get-PublicFolderDatabase $selDB)

        $dbMP = $false
        $logMP = $false

        if ($null -ne $mpvolumes) {
            foreach ($mp in $mpvolumes) {
                $mpname = (($mp.name).substring(0, $mp.name.length - 1))
                #if following mount point path exists in database path use deviceID in diskshadow config file

                if ($getDB.EdbFilePath.pathname.ToString().ToLower().StartsWith($mpname.ToString().ToLower())) {
                    Write-Host " "
                    Write-Host "Mount point: $($mp.name) in use for database path: "
                    "The current database path is: " + $getDB.EdbFilePath.pathname
                    Write-Host "adding deviceID to file: "
                    $dbEdbVol = $mp.deviceId
                    Write-Host $dbvol

                    #add device ID to array
                    $deviceID1 = $mp.DeviceID
                    $dbMP = $true
                }

                #if following mount point path exists in log path use deviceID in diskshadow config file
                if ($getDB.LogFolderPath.pathname.ToString().ToLower().contains($mpname.ToString().ToLower())) {
                    Write-Host " "
                    Write-Host "Mount point: $($vol.name) in use for log path: "
                    "The log folder path of selected database is: " + $getDB.LogFolderPath.pathname
                    Write-Host "adding deviceID to file "
                    $dbLogVol = $mp.deviceId
                    Write-Host $dblogvol

                    $deviceID2 = $mp.DeviceID
                    $logMP = $true
                }
            }
            $deviceIDs = $deviceID1, $deviceID2
        }
    }

    if ($dbMP -eq $false) {

        $dbEdbVol = ($getDB.EdbFilePath.pathname).substring(0, 2)
        "The selected database path is '" + $getDB.EdbFilePath.pathname + "' so adding volume $dbEdbVol to backup scope"
        $deviceID1 = $dbEdbVol
    }

    if ($logMP -eq $false) {
        $dbLogVol = ($getDB.LogFolderPath.pathname).substring(0, 2)
        $nl
        "The selected database log folder path is '" + $getDB.LogFolderPath.pathname + "' so adding volume $dbLogVol to backup scope"
        $deviceID2 = $dbLogVol
    }

    # Here is where we start adding the appropriate volumes or mountpoints to the diskshadow config file
    # We make sure that we add only one Logical volume when we detect the EDB and log files
    # are on the same volume

    $nl
    $deviceIDs = $deviceID1, $deviceID2
    $comp = [string]::Compare($deviceID1, $deviceID2, $True)
    if ($comp -eq 0) {
        $dID = $deviceIDs[0]
        Write-Debug -Message ('$dID = ' + $dID.ToString())
        Write-Debug "When the database and log files are on the same volume, we add the volume only once"
        if ($dID.length -gt "2") {
            $addVol = "add volume $dID alias vss_test_" + ($dID).tostring().substring(11, 8)
            Write-Host $addVol
            Out-DHSFile $addVol
        } else {
            $addVol = "add volume $dID alias vss_test_" + ($dID).tostring().substring(0, 1)
            Write-Host $addVol
            Out-DHSFile $addVol
        }
    } else {
        Write-Host " "
        foreach ($device in $deviceIDs) {
            if ($device.length -gt "2") {
                Write-Host "Adding the Mount Point for DSH file"
                $addVol = "add volume $device alias vss_test_" + ($device).tostring().substring(11, 8)
                Write-Host $addVol
                Out-DHSFile $addVol
            } else {
                Write-Host "Adding the volume for DSH file"
                $addVol = "add volume $device alias vss_test_" + ($device).tostring().substring(0, 1)
                Write-Host $addVol
                Out-DHSFile $addVol
            }
        }
    }
    Out-DHSFile "create"
    Out-DHSFile " "
    $nl
    Get-Date
    Write-Host "Getting drive letters for exposing backup snapshot" -ForegroundColor Green
    Write-Host "--------------------------------------------------------------------------------------------------------------"

    # check to see if the drives are the same for both database and logs
    # if the same volume is used, only one drive letter is needed for exposure
    # if two volumes are used, two drive letters are needed

    $matchCondition = "^[a-z]:$"
    Write-Debug $matchCondition

    if ($dbEdbVol -eq $dbLogVol) {
        $nl
        "Since the same volume is used for this database's EDB and logs, we only need a single drive"
        "letter to expose the backup snapshot."
        $nl

        do {
            Write-Host "Enter an unused drive letter with colon (e.g. X:) to expose the snapshot" -ForegroundColor Yellow -NoNewline
            $script:dbsnapvol = Read-Host " "
            if ($dbsnapvol -notmatch $matchCondition) {
                Write-Host "Your input was not acceptable. Please use a single letter and colon, e.g. X:" -ForegroundColor red
            }
        } while ($dbsnapvol -notmatch $matchCondition)
    } else {
        $nl
        "Since different volumes are used for this database's EDB and logs, we need two drive"
        "letters to expose the backup snapshot."
        $nl

        do {
            Write-Host "Enter an unused drive letter with colon (e.g. X:) to expose the DATABASE volume" -ForegroundColor Yellow -NoNewline
            $script:dbsnapvol = Read-Host " "
            if ($dbsnapvol -notmatch $matchCondition) {
                Write-Host "Your input was not acceptable. Please use a single letter and colon, e.g. X:" -ForegroundColor red
            }
        } while ($dbsnapvol -notmatch $matchCondition)

        do {
            Write-Host "Enter an unused drive letter with colon (e.g. Y:) to expose the LOG volume" -ForegroundColor Yellow -NoNewline
            $script:logsnapvol = Read-Host " "
            if ($logsnapvol -notmatch $matchCondition) {
                Write-Host "Your input was not acceptable. Please use a single letter and colon, e.g. Y:" -ForegroundColor red
            }
            if ($logsnapvol -eq $dbsnapvol) {
                Write-Host "You must choose a different drive letter than the one chosen to expose the DATABASE volume." -ForegroundColor red
            }
        } while (($logsnapvol -notmatch $matchCondition) -or ($logsnapvol -eq $dbsnapvol))

        $nl
    }

    Write-Debug "dbsnapvol: $dbsnapvol | logsnapvol: $logsnapvol"

    # expose the drives
    # if volumes are the same only one entry is needed
    if ($dbEdbVol -eq $dbLogVol) {
        if ($dbEdbVol.length -gt "2") {
            $dbvolstr = "expose %vss_test_" + ($dbEdbVol).substring(11, 8) + "% $dbsnapvol"
            Out-DHSFile $dbvolstr
        } else {
            $dbvolstr = "expose %vss_test_" + ($dbEdbVol).substring(0, 1) + "% $dbsnapvol"
            Out-DHSFile $dbvolstr
        }
    } else {
        # volumes are different, getting both
        # if mountpoint use first part of string, if not use first letter
        if ($dbEdbVol.length -gt "2") {
            $dbvolstr = "expose %vss_test_" + ($dbEdbVol).substring(11, 8) + "% $dbsnapvol"
            Out-DHSFile $dbvolstr
        } else {
            $dbvolstr = "expose %vss_test_" + ($dbEdbVol).substring(0, 1) + "% $dbsnapvol"
            Out-DHSFile $dbvolstr
        }

        # if mountpoint use first part of string, if not use first letter
        if ($dbLogVol.length -gt "2") {
            $logvolstr = "expose %vss_test_" + ($dbLogVol).substring(11, 8) + "% $logsnapvol"
            Out-DHSFile $logvolstr
        } else {
            $logvolstr = "expose %vss_test_" + ($dbLogVol).substring(0, 1) + "% $logsnapvol"
            Out-DHSFile $logvolstr
        }
    }

    # ending data of file
    Out-DHSFile "end backup"
}

function Invoke-DiskShadow {
    Write-Host " " $nl
    Get-Date
    Write-Host "Starting DiskShadow copy of Exchange database: $selDB" -ForegroundColor Green $nl
    Write-Host "--------------------------------------------------------------------------------------------------------------"
    " "
    Write-Host "Running the following command:" $nl
    Write-Host "`"C:\Windows\System32\diskshadow.exe /s $path\diskshadow.dsh /l $path\diskshadow.log`"" $nl
    Write-Host " "

    #in case the $path and the script location is different we need to change location into the $path directory to get the results to work as expected.
    try {
        $here = (Get-Location).Path
        Set-Location $path
        diskshadow.exe /s $path\diskshadow.dsh /l $path\diskshadow.log
    } finally {
        Set-Location $here
    }
}

function Invoke-RemoveExposedDrives {

    function Out-removeDHSFile {
        param ([string]$fileline)
        $fileline | Out-File -FilePath "$path\removeSnapshot.dsh" -Encoding ASCII -Append
    }

    " "
    Get-Date
    Write-Host "Diskshadow Snapshots" -ForegroundColor Green $nl
    Write-Host "--------------------------------------------------------------------------------------------------------------"
    " "
    Write-Host " "
    if ($null -eq $logsnapvol) {
        $exposedDrives = $dbsnapvol
    } else {
        $exposedDrives = $dbsnapvol.ToString() + " and " + $logsnapvol.ToString()
    }
    "If the snapshot was successful, the snapshot should be exposed as drive(s) $exposedDrives."
    "You should be able to see and navigate the snapshot with File Explorer. How would you like to proceed?"
    Write-Host " "
    Write-Host "NOTE: It is recommended to wait a few minutes to allow truncation to possibly occur before moving past this point." -ForegroundColor Cyan
    Write-Host "      This allows time for the logs that are automatically collected to include the window for the truncation to occur." -ForegroundColor Cyan
    Write-Host
    Write-Host "When ready, choose from the options below:" -ForegroundColor Yellow
    " "
    Write-Host "  1. Remove exposed snapshot now"
    Write-Host "  2. Keep snapshot exposed"
    Write-Host " "
    Write-Warning "Selecting option 1 will permanently delete the snapshot created, i.e. your backup will be deleted."
    " "
    $matchCondition = "^[1-2]$"
    Write-Debug "matchCondition: $matchCondition"
    do {
        Write-Host "Selection" -ForegroundColor Yellow -NoNewline
        $removeExpose = Read-Host " "
        if ($removeExpose -notmatch $matchCondition) {
            Write-Host "Error! Please choose a valid option." -ForegroundColor red
        }
    } while ($removeExpose -notmatch $matchCondition)

    $unexposeCommand = "delete shadows exposed $dbsnapvol"
    if ($null -ne $logsnapvol) {
        $unexposeCommand += $nl + "delete shadows exposed $logsnapvol"
    }

    if ($removeExpose -eq "1") {
        New-Item -Path $path\removeSnapshot.dsh -type file -Force
        Out-removeDHSFile $unexposeCommand
        Out-removeDHSFile "exit"
        & 'C:\Windows\System32\diskshadow.exe' /s $path\removeSnapshot.dsh
    } elseif ($removeExpose -eq "2") {
        Write-Host "You can remove the snapshots at a later time using the diskshadow tool from a command prompt."
        Write-Host "Run diskshadow followed by these commands:"
        Write-Host $unexposeCommand
    }
}

function Get-CopyStatus {
    if ((($databases[$dbToBackup]).IsMailboxDatabase) -eq "True") {
        Get-Date
        Write-Host "Status of '$selDB' and its replicas (if any)" -ForegroundColor Green $nl
        Write-Host "--------------------------------------------------------------------------------------------------------------"
        " "
        [array]$copyStatus = (Get-MailboxDatabaseCopyStatus -identity ($databases[$dbToBackup]).name)
        ($copyStatus | Format-List) | Out-File -FilePath "$path\copyStatus.txt"
        for ($i = 0; $i -lt ($copyStatus).length; $i++ ) {
            if (($copyStatus[$i].status -eq "Healthy") -or ($copyStatus[$i].status -eq "Mounted")) {
                Write-Host "$($copyStatus[$i].name) is $($copyStatus[$i].status)"
            } else {
                Write-Host "$($copyStatus[$i].name) is $($copyStatus[$i].status)"
                Write-Host "One of the copies of the selected database is not healthy. Please run backup after ensuring that the database copy is healthy" -ForegroundColor Yellow
                exit
            }
        }
    } else {
        Write-Host "Not checking database copy status since the selected database is a Public Folder Database..."
    }
    " "
}

function Get-Databases {
    Get-Date
    Write-Host "Getting databases on server: $serverName" -ForegroundColor Green $nl
    Write-Host "--------------------------------------------------------------------------------------------------------------"
    " "

    [array]$script:databases = Get-MailboxDatabase -server $serverName -status
    if ($null -ne (Get-PublicFolderDatabase -Server $serverName)) {
        $script:databases += Get-PublicFolderDatabase -server $serverName -status
    }

    #write-host "Database Name:`t`t Mounted: `t`t Mounted On Server:" -foregroundcolor Yellow $nl
    $script:dbID = 0

    foreach ($script:db in $databases) {
        $script:db | Add-Member NoteProperty Number $dbID
        $dbID++
    }

    $script:databases | Format-Table Number, Name, Mounted, Server -AutoSize | Out-String

    Write-Host " " $nl
}

function Get-DBtoBackup {
    $maxDbIndexRange = $script:databases.length - 1
    $matchCondition = "^([0-9]|[1-9][0-9])$"
    Write-Debug "matchCondition: $matchCondition"
    do {
        Write-Host "Select the number of the database to backup" -ForegroundColor Yellow -NoNewline;
        $script:dbToBackup = Read-Host " "

        if ($script:dbToBackup -notmatch $matchCondition -or [int]$script:dbToBackup -gt $maxDbIndexRange) {
            Write-Host "Error! Please select a valid option!" -ForegroundColor Red
        }
    } while ($script:dbToBackup -notmatch $matchCondition -or [int]$script:dbToBackup -gt $maxDbIndexRange) # notmatch is case-insensitive

    if ((($databases[$dbToBackup]).IsMailboxDatabase) -eq "True") {

        $script:dbGuid = (Get-MailboxDatabase ($databases[$dbToBackup])).guid
        $script:selDB = (Get-MailboxDatabase ($databases[$dbToBackup])).name
        " "
        "The database guid for '$selDB' is: $dbGuid"
        " "
        $script:dbMountedOn = (Get-MailboxDatabase ($databases[$dbToBackup])).server.name
    } else {
        $script:dbGuid = (Get-PublicFolderDatabase ($databases[$dbToBackup])).guid
        $script:selDB = (Get-PublicFolderDatabase ($databases[$dbToBackup])).name
        "The database guid for '$selDB' is: $dbGuid"
        " "
        $script:dbMountedOn = (Get-PublicFolderDatabase ($databases[$dbToBackup])).server.name
    }
    Write-Host "The database is mounted on server: $dbMountedOn $nl"

    if ($dbMountedOn -eq "$serverName") {
        $script:dbStatus = "active"
    } else {
        $script:dbStatus = "passive"
    }
}

function Get-ExchangeVersion {
    Get-Date
    Write-Host "Verifying Exchange version..." -ForegroundColor Green $nl
    Write-Host "--------------------------------------------------------------------------------------------------------------"
    " "
    $script:exchVer = (Get-ExchangeServer $serverName).AdminDisplayVersion
    $exchVerMajor = $exchVer.major
    $exchVerMinor = $exchVer.minor

    switch ($exchVerMajor) {
        "14" {
            $script:exchVer = "2010"
        }
        "15" {
            switch ($exchVerMinor) {
                "0" {
                    $script:exchVer = "2013"
                }
                "1" {
                    $script:exchVer = "2016"
                }
                "2" {
                    $script:exchVer = "2019"
                }
            }
        }

        default {
            Write-Host "This script is only for Exchange 2013, 2016, and 2019 servers." -ForegroundColor red $nl
            exit
        }
    }

    Write-Host "$serverName is an Exchange $exchVer server. $nl"

    if ($exchVer -eq "2010") {
        Write-Host "This script no longer supports Exchange 2010."
        exit
    }
}


function Get-WindowsEventLogs {

    function Get-WindowEventsPerServer {
        param(
            [string]$CompunterName
        )
        "Getting application log events..."
        Get-WinEvent -FilterHashtable @{LogName = "Application"; StartTime = $startTime } -ComputerName $CompunterName | Export-Clixml $path\events-$CompunterName-App.xml
        "Getting system log events..."
        Get-WinEvent -FilterHashtable @{LogName = "System"; StartTime = $startTime } -ComputerName $CompunterName | Export-Clixml $path\events-$CompunterName-System.xml
        "Getting TruncationDebug log events..."
        Get-WinEvent -FilterHashtable @{LogName = "Microsoft-Exchange-HighAvailability/TruncationDebug"; StartTime = $startTime } -ComputerName $CompunterName -ErrorAction SilentlyContinue | Export-Clixml $path\events-$CompunterName-TruncationDebug.xml
    }
    " "
    Get-Date
    Write-Host "Getting events from the application and system logs since the script's start time of ($startInfo)" -ForegroundColor Green $nl
    Write-Host "--------------------------------------------------------------------------------------------------------------"
    " "
    "Getting application log events..."
    Get-EventLog -LogName Application -After $startInfo | Export-Clixml $path\events-App.xml
    "Getting system log events..."
    Get-EventLog -LogName System -After $startInfo | Export-Clixml $path\events-System.xml
    "Getting events complete!"
}

function Get-VSSWritersAfter {
    " "
    Get-Date
    Write-Host "Checking VSS Writer Status: (after backup)" -ForegroundColor Green $nl
    Write-Host "--------------------------------------------------------------------------------------------------------------"
    " "
    " "
    $writers = (vssadmin list writers)
    $writers > $path\vssWritersAfter.txt

    foreach ($line in $writers) {
        if ($line -like "Writer name:*") {
            "$line"
        } elseif ($line -like "   State:*") {
            "$line" + $nl
        }
    }
}

function Get-VSSWritersBefore {
    " "
    Get-Date
    Write-Host "Checking VSS Writer Status: (All Writers must be in a Stable state before running this script)" -ForegroundColor Green $nl
    Write-Host "--------------------------------------------------------------------------------------------------------------"
    " "
    $writers = (vssadmin list writers)
    $writers > $path\vssWritersBefore.txt
    $exchangeWriter = $false

    foreach ($line in $writers) {

        if ($line -like "Writer name:*") {
            "$line"

            if ($line.Contains("Microsoft Exchange Writer")) {
                $exchangeWriter = $true
            }
        } elseif ($line -like "   State:*") {

            if ($line -ne "   State: [1] Stable") {
                $nl
                Write-Host "!!!!!!!!!!!!!!!!!!!!!!!!!!   WARNING   !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!" -ForegroundColor red
                $nl
                Write-Host "One or more writers are NOT in a 'Stable' state, STOPPING SCRIPT." -ForegroundColor red
                $nl
                Write-Host "Review the vssWritersBefore.txt file in '$path' for more information." -ForegroundColor Red
                Write-Host "You can also use an Exchange Management Shell or a Command Prompt to run: 'vssadmin list writers'" -ForegroundColor red
                $nl
                Write-Host "!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!" -ForegroundColor red
                $nl
                exit
            } else {
                "$line" + $nl
            }
        }
    }
    " " + $nl

    if (!$exchangeWriter) {

        #Check for possible COM security issue.
        $oleKey = "HKLM:\SOFTWARE\Microsoft\Ole"
        $dcomKey = "HKLM:\SOFTWARE\Policies\Microsoft\Windows NT\DCOM"

        $possibleDcomPermissionIssue = ((Test-Path $oleKey) -and
            ($null -ne (Get-ItemProperty $oleKey).DefaultAccessPermission)) -or
        ((Test-Path $dcomKey) -and
            ($null -ne (Get-ItemProperty $dcomKey).MachineAccessRestriction))

        Write-Host "!!!!!!!!!!!!!!!!!!!!!!!!!!   WARNING   !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!" -ForegroundColor red
        Write-Host "Microsoft Exchange Writer not present on server. Unable to preform proper backups on the server."  -ForegroundColor Red
        Write-Host

        if ($possibleDcomPermissionIssue) {
            Write-Host " - Recommend to verify local Administrators group applied to COM+ Security settings: https://aka.ms/VSSTester-COMSecurity" -ForegroundColor Cyan
        }

        Write-Host " - Recommend to restart MSExchangeRepl service to see if the writer comes back. If it doesn't, review the application logs for any events to determine why." -ForegroundColor Cyan
        Write-Host " --- Look for Event ID 2003 in the application logs to verify that all internal components come online. If you see this event, try to use PSExec.exe to start a cmd.exe as the SYSTEM account and run 'vssadmin list writers'" -ForegroundColor Cyan
        Write-Host " --- If you find the Microsoft Exchange Writer, then we have a permissions issue on the computer that is preventing normal user accounts from finding all the writers." -ForegroundColor Cyan
        Write-Host " - If still not able to determine why, need to have a Microsoft Engineer review ExTrace with Cluster.Replay tags of the MSExchangeRepl service starting up." -ForegroundColor Cyan
        Write-Host
        Write-Host "Stopping Script"
        exit
    }
}

function Invoke-CreateExTRATracingConfig {

    function Out-ExTRAConfigFile {
        param ([string]$fileline)
        $fileline | Out-File -FilePath "C:\EnabledTraces.Config" -Encoding ASCII -Append
    }

    " "
    Get-Date
    Write-Host "Enabling ExTRA Tracing..." -ForegroundColor Green $nl
    Write-Host "--------------------------------------------------------------------------------------------------------------"
    " "
    New-Item -Path "C:\EnabledTraces.Config" -type file -Force

    Out-ExTRAConfigFile "TraceLevels:Debug,Warning,Error,Fatal,Info,Performance,Function,Pfd"
    Out-ExTRAConfigFile "ManagedStore.PhysicalAccess:JetBackup,JetRestore,JetEventlog,SnapshotOperation"
    Out-ExTRAConfigFile "Cluster.Replay:LogTruncater,ReplayApi,ReplicaInstance,ReplicaVssWriterInterop"
    Out-ExTRAConfigFile "ManagedStore.HA:BlockModeSender,Eseback"
    Out-ExTRAConfigFile "FilteredTracing:No"
    Out-ExTRAConfigFile "InMemoryTracing:No"
    " "
    Write-Debug "ExTRA trace config file created successfully"
}

function Invoke-DisableDiagnosticsLogging {

    Write-Host " "  $nl
    Get-Date
    Write-Host "Disabling Diagnostics Logging..." -ForegroundColor green $nl
    Write-Host "--------------------------------------------------------------------------------------------------------------"
    " "
    Set-EventLogLevel 'MSExchange Repl\Service' -level lowest
    $disgetReplSvc = Get-EventLogLevel 'MSExchange Repl\Service'
    Write-Host "$($disgetReplSvc.Identity) - $($disgetReplSvc.EventLevel) $nl"

    Set-EventLogLevel 'MSExchange Repl\Exchange VSS Writer' -level lowest
    $disgetReplVSSWriter = Get-EventLogLevel 'MSExchange Repl\Exchange VSS Writer'
    Write-Host "$($disgetReplVSSWriter.Identity) - $($disgetReplVSSWriter.EventLevel) $nl"
}

function Invoke-DisableExTRATracing {
    " "
    Get-Date
    Write-Host "Disabling ExTRA Tracing..." -ForegroundColor Green $nl
    Write-Host "--------------------------------------------------------------------------------------------------------------"
    " "
    if ($dbMountedOn -eq "$serverName") {
        #stop active copy
        Write-Host " "
        "Stopping Exchange Trace data collector on $serverName..."
        logman stop vssTester -s $serverName
        "Deleting Exchange Trace data collector on $serverName..."
        logman delete vssTester -s $serverName
        " "
    } else {
        #stop passive copy
        "Stopping Exchange Trace data collector on $serverName..."
        logman stop vssTester-Passive -s $serverName
        "Deleting Exchange Trace data collector on $serverName..."
        logman delete vssTester-Passive -s $serverName
        #stop active copy
        "Stopping Exchange Trace data collector on $dbMountedOn..."
        logman stop vssTester-Active -s $dbMountedOn
        "Deleting Exchange Trace data collector on $dbMountedOn..."
        logman delete vssTester-Active -s $dbMountedOn
        " "
        "Moving ETL file from $dbMountedOn to $serverName..."
        " "
        $etlPath = $path -replace ":\\", "$\"
        Move-Item "\\$dbMountedOn\$etlPath\vsstester-active_000001.etl" "\\$servername\$etlPath\vsstester-active_000001.etl" -Force
    }
}

function Invoke-DisableVSSTracing {
    " "
    Get-Date
    Write-Host "Disabling VSS Tracing..." -ForegroundColor Green $nl
    Write-Host "--------------------------------------------------------------------------------------------------------------"
    " "
    logman stop vss -ets
    " "
}

function Invoke-EnableDiagnosticsLogging {
    " "
    Get-Date
    Write-Host "Enabling Diagnostics Logging..." -ForegroundColor green $nl
    Write-Host "--------------------------------------------------------------------------------------------------------------"
    " "
    Set-EventLogLevel 'MSExchange Repl\Service' -level expert
    $getReplSvc = Get-EventLogLevel 'MSExchange Repl\Service'
    Write-Host "$($getReplSvc.Identity) - $($getReplSvc.EventLevel) $nl"

    Set-EventLogLevel 'MSExchange Repl\Exchange VSS Writer' -level expert
    $getReplVSSWriter = Get-EventLogLevel 'MSExchange Repl\Exchange VSS Writer'
    Write-Host "$($getReplVSSWriter.Identity)  - $($getReplVSSWriter.EventLevel)  $nl"
}

function Invoke-EnableExTRATracing {

    function Invoke-ExtraTracingCreate {
        param(
            [string]$ComputerName,
            [string]$LogmanName
        )
        logman create trace $LogmanName -p '{79bb49e6-2a2c-46e4-9167-fa122525d540}' -o $path\$LogmanName.etl -ow -s $ComputerName -mode globalsequence

        if ($LASTEXITCODE) {
            Write-Host "Exchange Trace data Collector set already created. Removing it and trying again"
            logman delete $LogmanName -s $ComputerName

            logman create trace $LogmanName -p '{79bb49e6-2a2c-46e4-9167-fa122525d540}' -o $path\$LogmanName.etl -ow -s $ComputerName -mode globalsequence
        }

        if ($LASTEXITCODE) {
            Write-Host "Failed to create the extra trace. Stopping the VSSTester Script" -ForegroundColor Red
            exit
        }
    }

    #active server, only get tracing from active node
    if ($dbMountedOn -eq $serverName) {
        " "
        "Creating Exchange Trace data collector set..."
        Invoke-ExtraTracingCreate -ComputerName $serverName -LogmanName "VSSTester"
        "Starting Exchange Trace data collector..."
        logman start VSSTester

        if ($LASTEXITCODE) {
            Write-Host "Failed to start the extra trace. Stopping the VSSTester Script" -ForegroundColor Red
            exit
        }
        " "
    } else {
        #passive server, get tracing from both active and passive nodes
        " "
        "Copying the ExTRA config file 'EnabledTraces.config' file to $dbMountedOn..."
        #copy enabledtraces.config from current passive copy to active copy server
        Copy-Item "c:\EnabledTraces.Config" "\\$dbMountedOn\c$\enabledtraces.config" -Force

        #create trace on passive copy
        "Creating Exchange Trace data collector set on $serverName..."
        Invoke-ExtraTracingCreate -ComputerName $serverName -LogmanName "VSSTester-Passive"
        #create trace on active copy
        "Creating Exchange Trace data collector set on $dbMountedOn..."
        Invoke-ExtraTracingCreate -ComputerName $dbMountedOn -LogmanName "VSSTester-Active"
        #start trace on passive copy
        "Starting Exchange Trace data collector on $serverName..."
        logman start VSSTester-Passive -s $serverName

        if ($LASTEXITCODE) {
            Write-Host "Failed to start the extra trace. Stopping the VSSTester Script" -ForegroundColor Red
            exit
        }
        #start trace on active copy
        "Starting Exchange Trace data collector on $dbMountedOn..."
        logman start VSSTester-Active -s $dbMountedOn

        if ($LASTEXITCODE) {
            Write-Host "Failed to start the extra trace. Stopping the VSSTester Script" -ForegroundColor Red
            exit
        }
        " "
    }

    Write-Debug "ExTRA trace started successfully"
}

function Invoke-EnableVSSTracing {
    " "
    Get-Date
    Write-Host "Enabling VSS Tracing..." -ForegroundColor Green $nl
    Write-Host "--------------------------------------------------------------------------------------------------------------"
    " "
    logman start vss -o $path\vss.etl -ets -p "{9138500e-3648-4edb-aa4c-859e9f7b7c38}" 0xfff 255
}

function Main {
    $updateInfo = Get-ScriptUpdateAvailable
    if ($updateInfo.UpdateFound) {
        Write-Warning "An update is available for this script. Current: $($updateInfo.CurrentVersion) Latest: $($updateInfo.LatestVersion)"
        Write-Warning "Please download the latest: https://microsoft.github.io/CSS-Exchange/Databases/VSSTester/"
    }

    # if a transcript is running, we need to stop it as this script will start its own
    try {
        Stop-Transcript | Out-Null
    } catch [System.InvalidOperationException] { }

    Write-Host "****************************************************************************************"
    Write-Host "****************************************************************************************"
    Write-Host "**                                                                                    **" -BackgroundColor DarkMagenta
    Write-Host "**                 VSSTESTER SCRIPT (for Exchange 2013, 2016, 2019)                   **" -ForegroundColor Cyan -BackgroundColor DarkMagenta
    Write-Host "**                                                                                    **" -BackgroundColor DarkMagenta
    Write-Host "****************************************************************************************"
    Write-Host "****************************************************************************************"

    $Script:LocalExchangeShell = Confirm-ExchangeShell -Identity $env:COMPUTERNAME

    if (!$Script:LocalExchangeShell.ShellLoaded) {
        Write-Host "Failed to load Exchange Shell. Stopping the script."
        exit
    }

    if ($Script:LocalExchangeShell.RemoteShell -or
        $Script:LocalExchangeShell.ToolsOnly) {
        Write-Host "Can't run this script from a non Exchange Server."
        exit
    }

    #newLine shortcut
    $script:nl = "`r`n"
    $nl

    $script:serverName = $env:COMPUTERNAME

    #start time
    $Script:startInfo = Get-Date
    Get-Date

    if ($DebugPreference -ne 'SilentlyContinue') {
        $nl
        Write-Host 'This script is running in DEBUG mode since $DebugPreference is not set to SilentlyContinue.' -ForegroundColor Red
    }

    $nl
    Write-Host "Please select the operation you would like to perform from the following options:" -ForegroundColor Green
    $nl
    Write-Host "  1. " -ForegroundColor Yellow -NoNewline; Write-Host "Test backup using built-in Diskshadow"
    Write-Host "  2. " -ForegroundColor Yellow -NoNewline; Write-Host "Enable logging to troubleshoot backup issues"
    $nl

    $matchCondition = "^[1|2]$"
    Write-Debug "matchCondition: $matchCondition"
    do {
        Write-Host "Selection: " -ForegroundColor Yellow -NoNewline;
        $Selection = Read-Host
        if ($Selection -notmatch $matchCondition) {
            Write-Host "Error! Please select a valid option!" -ForegroundColor Red
        }
    }
    while ($Selection -notmatch $matchCondition)


    try {

        $nl
        Write-Host "Please specify a directory other than root of a volume to save the configuration and output files." -ForegroundColor Green

        $pathExists = $false

        # get path, ensuring it exists
        do {
            Write-Host "Directory path (e.g. C:\temp): " -ForegroundColor Yellow -NoNewline
            $script:path = Read-Host
            Write-Debug "path: $path"
            try {
                $pathExists = Test-Path -Path "$path"
            } catch { }
            Write-Debug "pathExists: $pathExists"
            if ($pathExists -ne $true) {
                Write-Host "Error! The path does not exist. Please enter a valid path." -ForegroundColor red
            }
        } while ($pathExists -ne $true)

        $nl
        Get-Date
        Write-Host "Starting transcript..." -ForegroundColor Green $nl
        Write-Host "--------------------------------------------------------------------------------------------------------------"

        Start-Transcript -Path "$($script:path)\vssTranscript.log"
        $nl

        if ($Selection -eq 1) {
            Get-ExchangeVersion
            Get-VSSWritersBefore
            Get-Databases
            Get-DBtoBackup
            Get-CopyStatus
            Invoke-CreateDiskShadowFile #---
            Invoke-EnableDiagnosticsLogging
            Invoke-EnableVSSTracing
            Invoke-CreateExTRATracingConfig
            Invoke-EnableExTRATracing
            Invoke-DiskShadow #---
            Get-VSSWritersAfter
            Invoke-RemoveExposedDrives #---
            Invoke-DisableExTRATracing
            Invoke-DisableDiagnosticsLogging
            Invoke-DisableVSSTracing
            Get-WindowsEventLogs
        } elseif ($Selection -eq 2) {
            Get-ExchangeVersion
            Get-VSSWritersBefore
            Get-Databases
            Get-DBtoBackup
            Get-CopyStatus
            Invoke-EnableDiagnosticsLogging
            Invoke-EnableVSSTracing
            Invoke-CreateExTRATracingConfig
            Invoke-EnableExTRATracing

            #Here is where we wait for the end user to perform the backup using the backup software and then come back to the script to press "Enter", thereby stopping data collection
            Get-Date
            Write-Host "Data Collection" -ForegroundColor green $nl
            Write-Host "--------------------------------------------------------------------------------------------------------------"
            " "
            Write-Host "Data collection is now enabled." -ForegroundColor Yellow
            Write-Host "Please start your backup using the third party software so the script can record the diagnostic data." -ForegroundColor Yellow
            Write-Host "When the backup is COMPLETE, use the <Enter> key to terminate data collection..." -ForegroundColor Yellow -NoNewline
            Read-Host

            Invoke-DisableExTRATracing
            Invoke-DisableDiagnosticsLogging
            Invoke-DisableVSSTracing
            Get-VSSWritersAfter
            Get-WindowsEventLogs
        }
    } finally {
        # always stop our transcript at end of script's execution
        # we catch a failure here if we try to stop a transcript that's not running
        try {
            " " + $nl
            Get-Date
            Write-Host "Stopping transcript log..." -ForegroundColor Green $nl
            Write-Host "--------------------------------------------------------------------------------------------------------------"
            " "
            Stop-Transcript
            " " + $nl
            do {
                Write-Host
                $continue = Read-Host "Please use the <Enter> key to exit..."
            }
            while ($null -notmatch $continue)
            exit
        } catch { }
    }
}

try {
    Clear-Host
    Main
} catch { } finally { }

# SIG # Begin signature block
# MIInvAYJKoZIhvcNAQcCoIInrTCCJ6kCAQExDzANBglghkgBZQMEAgEFADB5Bgor
# BgEEAYI3AgEEoGswaTA0BgorBgEEAYI3AgEeMCYCAwEAAAQQH8w7YFlLCE63JNLG
# KX7zUQIBAAIBAAIBAAIBAAIBADAxMA0GCWCGSAFlAwQCAQUABCCP9i5uK5OlRQnR
# fNuGbMsS1IxH5+qW/PhQfd6NY2ftWqCCDYUwggYDMIID66ADAgECAhMzAAACzfNk
# v/jUTF1RAAAAAALNMA0GCSqGSIb3DQEBCwUAMH4xCzAJBgNVBAYTAlVTMRMwEQYD
# VQQIEwpXYXNoaW5ndG9uMRAwDgYDVQQHEwdSZWRtb25kMR4wHAYDVQQKExVNaWNy
# b3NvZnQgQ29ycG9yYXRpb24xKDAmBgNVBAMTH01pY3Jvc29mdCBDb2RlIFNpZ25p
# bmcgUENBIDIwMTEwHhcNMjIwNTEyMjA0NjAyWhcNMjMwNTExMjA0NjAyWjB0MQsw
# CQYDVQQGEwJVUzETMBEGA1UECBMKV2FzaGluZ3RvbjEQMA4GA1UEBxMHUmVkbW9u
# ZDEeMBwGA1UEChMVTWljcm9zb2Z0IENvcnBvcmF0aW9uMR4wHAYDVQQDExVNaWNy
# b3NvZnQgQ29ycG9yYXRpb24wggEiMA0GCSqGSIb3DQEBAQUAA4IBDwAwggEKAoIB
# AQDrIzsY62MmKrzergm7Ucnu+DuSHdgzRZVCIGi9CalFrhwtiK+3FIDzlOYbs/zz
# HwuLC3hir55wVgHoaC4liQwQ60wVyR17EZPa4BQ28C5ARlxqftdp3H8RrXWbVyvQ
# aUnBQVZM73XDyGV1oUPZGHGWtgdqtBUd60VjnFPICSf8pnFiit6hvSxH5IVWI0iO
# nfqdXYoPWUtVUMmVqW1yBX0NtbQlSHIU6hlPvo9/uqKvkjFUFA2LbC9AWQbJmH+1
# uM0l4nDSKfCqccvdI5l3zjEk9yUSUmh1IQhDFn+5SL2JmnCF0jZEZ4f5HE7ykDP+
# oiA3Q+fhKCseg+0aEHi+DRPZAgMBAAGjggGCMIIBfjAfBgNVHSUEGDAWBgorBgEE
# AYI3TAgBBggrBgEFBQcDAzAdBgNVHQ4EFgQU0WymH4CP7s1+yQktEwbcLQuR9Zww
# VAYDVR0RBE0wS6RJMEcxLTArBgNVBAsTJE1pY3Jvc29mdCBJcmVsYW5kIE9wZXJh
# dGlvbnMgTGltaXRlZDEWMBQGA1UEBRMNMjMwMDEyKzQ3MDUzMDAfBgNVHSMEGDAW
# gBRIbmTlUAXTgqoXNzcitW2oynUClTBUBgNVHR8ETTBLMEmgR6BFhkNodHRwOi8v
# d3d3Lm1pY3Jvc29mdC5jb20vcGtpb3BzL2NybC9NaWNDb2RTaWdQQ0EyMDExXzIw
# MTEtMDctMDguY3JsMGEGCCsGAQUFBwEBBFUwUzBRBggrBgEFBQcwAoZFaHR0cDov
# L3d3dy5taWNyb3NvZnQuY29tL3BraW9wcy9jZXJ0cy9NaWNDb2RTaWdQQ0EyMDEx
# XzIwMTEtMDctMDguY3J0MAwGA1UdEwEB/wQCMAAwDQYJKoZIhvcNAQELBQADggIB
# AE7LSuuNObCBWYuttxJAgilXJ92GpyV/fTiyXHZ/9LbzXs/MfKnPwRydlmA2ak0r
# GWLDFh89zAWHFI8t9JLwpd/VRoVE3+WyzTIskdbBnHbf1yjo/+0tpHlnroFJdcDS
# MIsH+T7z3ClY+6WnjSTetpg1Y/pLOLXZpZjYeXQiFwo9G5lzUcSd8YVQNPQAGICl
# 2JRSaCNlzAdIFCF5PNKoXbJtEqDcPZ8oDrM9KdO7TqUE5VqeBe6DggY1sZYnQD+/
# LWlz5D0wCriNgGQ/TWWexMwwnEqlIwfkIcNFxo0QND/6Ya9DTAUykk2SKGSPt0kL
# tHxNEn2GJvcNtfohVY/b0tuyF05eXE3cdtYZbeGoU1xQixPZAlTdtLmeFNly82uB
# VbybAZ4Ut18F//UrugVQ9UUdK1uYmc+2SdRQQCccKwXGOuYgZ1ULW2u5PyfWxzo4
# BR++53OB/tZXQpz4OkgBZeqs9YaYLFfKRlQHVtmQghFHzB5v/WFonxDVlvPxy2go
# a0u9Z+ZlIpvooZRvm6OtXxdAjMBcWBAsnBRr/Oj5s356EDdf2l/sLwLFYE61t+ME
# iNYdy0pXL6gN3DxTVf2qjJxXFkFfjjTisndudHsguEMk8mEtnvwo9fOSKT6oRHhM
# 9sZ4HTg/TTMjUljmN3mBYWAWI5ExdC1inuog0xrKmOWVMIIHejCCBWKgAwIBAgIK
# YQ6Q0gAAAAAAAzANBgkqhkiG9w0BAQsFADCBiDELMAkGA1UEBhMCVVMxEzARBgNV
# BAgTCldhc2hpbmd0b24xEDAOBgNVBAcTB1JlZG1vbmQxHjAcBgNVBAoTFU1pY3Jv
# c29mdCBDb3Jwb3JhdGlvbjEyMDAGA1UEAxMpTWljcm9zb2Z0IFJvb3QgQ2VydGlm
# aWNhdGUgQXV0aG9yaXR5IDIwMTEwHhcNMTEwNzA4MjA1OTA5WhcNMjYwNzA4MjEw
# OTA5WjB+MQswCQYDVQQGEwJVUzETMBEGA1UECBMKV2FzaGluZ3RvbjEQMA4GA1UE
# BxMHUmVkbW9uZDEeMBwGA1UEChMVTWljcm9zb2Z0IENvcnBvcmF0aW9uMSgwJgYD
# VQQDEx9NaWNyb3NvZnQgQ29kZSBTaWduaW5nIFBDQSAyMDExMIICIjANBgkqhkiG
# 9w0BAQEFAAOCAg8AMIICCgKCAgEAq/D6chAcLq3YbqqCEE00uvK2WCGfQhsqa+la
# UKq4BjgaBEm6f8MMHt03a8YS2AvwOMKZBrDIOdUBFDFC04kNeWSHfpRgJGyvnkmc
# 6Whe0t+bU7IKLMOv2akrrnoJr9eWWcpgGgXpZnboMlImEi/nqwhQz7NEt13YxC4D
# dato88tt8zpcoRb0RrrgOGSsbmQ1eKagYw8t00CT+OPeBw3VXHmlSSnnDb6gE3e+
# lD3v++MrWhAfTVYoonpy4BI6t0le2O3tQ5GD2Xuye4Yb2T6xjF3oiU+EGvKhL1nk
# kDstrjNYxbc+/jLTswM9sbKvkjh+0p2ALPVOVpEhNSXDOW5kf1O6nA+tGSOEy/S6
# A4aN91/w0FK/jJSHvMAhdCVfGCi2zCcoOCWYOUo2z3yxkq4cI6epZuxhH2rhKEmd
# X4jiJV3TIUs+UsS1Vz8kA/DRelsv1SPjcF0PUUZ3s/gA4bysAoJf28AVs70b1FVL
# 5zmhD+kjSbwYuER8ReTBw3J64HLnJN+/RpnF78IcV9uDjexNSTCnq47f7Fufr/zd
# sGbiwZeBe+3W7UvnSSmnEyimp31ngOaKYnhfsi+E11ecXL93KCjx7W3DKI8sj0A3
# T8HhhUSJxAlMxdSlQy90lfdu+HggWCwTXWCVmj5PM4TasIgX3p5O9JawvEagbJjS
# 4NaIjAsCAwEAAaOCAe0wggHpMBAGCSsGAQQBgjcVAQQDAgEAMB0GA1UdDgQWBBRI
# bmTlUAXTgqoXNzcitW2oynUClTAZBgkrBgEEAYI3FAIEDB4KAFMAdQBiAEMAQTAL
# BgNVHQ8EBAMCAYYwDwYDVR0TAQH/BAUwAwEB/zAfBgNVHSMEGDAWgBRyLToCMZBD
# uRQFTuHqp8cx0SOJNDBaBgNVHR8EUzBRME+gTaBLhklodHRwOi8vY3JsLm1pY3Jv
# c29mdC5jb20vcGtpL2NybC9wcm9kdWN0cy9NaWNSb29DZXJBdXQyMDExXzIwMTFf
# MDNfMjIuY3JsMF4GCCsGAQUFBwEBBFIwUDBOBggrBgEFBQcwAoZCaHR0cDovL3d3
# dy5taWNyb3NvZnQuY29tL3BraS9jZXJ0cy9NaWNSb29DZXJBdXQyMDExXzIwMTFf
# MDNfMjIuY3J0MIGfBgNVHSAEgZcwgZQwgZEGCSsGAQQBgjcuAzCBgzA/BggrBgEF
# BQcCARYzaHR0cDovL3d3dy5taWNyb3NvZnQuY29tL3BraW9wcy9kb2NzL3ByaW1h
# cnljcHMuaHRtMEAGCCsGAQUFBwICMDQeMiAdAEwAZQBnAGEAbABfAHAAbwBsAGkA
# YwB5AF8AcwB0AGEAdABlAG0AZQBuAHQALiAdMA0GCSqGSIb3DQEBCwUAA4ICAQBn
# 8oalmOBUeRou09h0ZyKbC5YR4WOSmUKWfdJ5DJDBZV8uLD74w3LRbYP+vj/oCso7
# v0epo/Np22O/IjWll11lhJB9i0ZQVdgMknzSGksc8zxCi1LQsP1r4z4HLimb5j0b
# pdS1HXeUOeLpZMlEPXh6I/MTfaaQdION9MsmAkYqwooQu6SpBQyb7Wj6aC6VoCo/
# KmtYSWMfCWluWpiW5IP0wI/zRive/DvQvTXvbiWu5a8n7dDd8w6vmSiXmE0OPQvy
# CInWH8MyGOLwxS3OW560STkKxgrCxq2u5bLZ2xWIUUVYODJxJxp/sfQn+N4sOiBp
# mLJZiWhub6e3dMNABQamASooPoI/E01mC8CzTfXhj38cbxV9Rad25UAqZaPDXVJi
# hsMdYzaXht/a8/jyFqGaJ+HNpZfQ7l1jQeNbB5yHPgZ3BtEGsXUfFL5hYbXw3MYb
# BL7fQccOKO7eZS/sl/ahXJbYANahRr1Z85elCUtIEJmAH9AAKcWxm6U/RXceNcbS
# oqKfenoi+kiVH6v7RyOA9Z74v2u3S5fi63V4GuzqN5l5GEv/1rMjaHXmr/r8i+sL
# gOppO6/8MO0ETI7f33VtY5E90Z1WTk+/gFcioXgRMiF670EKsT/7qMykXcGhiJtX
# cVZOSEXAQsmbdlsKgEhr/Xmfwb1tbWrJUnMTDXpQzTGCGY0wghmJAgEBMIGVMH4x
# CzAJBgNVBAYTAlVTMRMwEQYDVQQIEwpXYXNoaW5ndG9uMRAwDgYDVQQHEwdSZWRt
# b25kMR4wHAYDVQQKExVNaWNyb3NvZnQgQ29ycG9yYXRpb24xKDAmBgNVBAMTH01p
# Y3Jvc29mdCBDb2RlIFNpZ25pbmcgUENBIDIwMTECEzMAAALN82S/+NRMXVEAAAAA
# As0wDQYJYIZIAWUDBAIBBQCggcYwGQYJKoZIhvcNAQkDMQwGCisGAQQBgjcCAQQw
# HAYKKwYBBAGCNwIBCzEOMAwGCisGAQQBgjcCARUwLwYJKoZIhvcNAQkEMSIEIAIn
# AwMvIGWvDVYtwjKtHlRbbU21uROoq2CvVn/mWmtjMFoGCisGAQQBgjcCAQwxTDBK
# oBqAGABDAFMAUwAgAEUAeABjAGgAYQBuAGcAZaEsgCpodHRwczovL2dpdGh1Yi5j
# b20vbWljcm9zb2Z0L0NTUy1FeGNoYW5nZSAwDQYJKoZIhvcNAQEBBQAEggEAHZ91
# veBjX/hKkTWvUnO81jZ8udtd5ejqOyl62FK+XsadQQroLirfZeCdRHEDAt8+SK4S
# /XmGf6ceV92FzBJ8dCJB1HRP8k2awgZ7XDt6cElAmIxi6RWg+497AJik3rlnE9hu
# 6e8zJAUVxMImwhCkPB34m97G04y2F5X6RH32bFf3UzW0363FdkkPtRbJblOx93KR
# yEl0unzzG/EnlRjPcpimz+73+tQsDVQvoitFrnySd1Jr6y+T37P9w341nypBKQ1h
# qr3xtNVDBetZGDol8XT2iPEVlOiPLMlZkAFqbGc7yqNB7bJCgB7ujJzb654GcW+T
# fRgCjNYwZiAgzKkfZKGCFv8wghb7BgorBgEEAYI3AwMBMYIW6zCCFucGCSqGSIb3
# DQEHAqCCFtgwghbUAgEDMQ8wDQYJYIZIAWUDBAIBBQAwggFQBgsqhkiG9w0BCRAB
# BKCCAT8EggE7MIIBNwIBAQYKKwYBBAGEWQoDATAxMA0GCWCGSAFlAwQCAQUABCDZ
# 9K4aWD5CcMNr5CjbYD0uqQpf8iYqcu4rkex3kiU+nwIGYv1r1AdsGBIyMDIyMDgz
# MTE2MTQwNi4yOFowBIACAfSggdCkgc0wgcoxCzAJBgNVBAYTAlVTMRMwEQYDVQQI
# EwpXYXNoaW5ndG9uMRAwDgYDVQQHEwdSZWRtb25kMR4wHAYDVQQKExVNaWNyb3Nv
# ZnQgQ29ycG9yYXRpb24xJTAjBgNVBAsTHE1pY3Jvc29mdCBBbWVyaWNhIE9wZXJh
# dGlvbnMxJjAkBgNVBAsTHVRoYWxlcyBUU1MgRVNOOkFFMkMtRTMyQi0xQUZDMSUw
# IwYDVQQDExxNaWNyb3NvZnQgVGltZS1TdGFtcCBTZXJ2aWNloIIRVzCCBwwwggT0
# oAMCAQICEzMAAAGWSVti4S/d908AAQAAAZYwDQYJKoZIhvcNAQELBQAwfDELMAkG
# A1UEBhMCVVMxEzARBgNVBAgTCldhc2hpbmd0b24xEDAOBgNVBAcTB1JlZG1vbmQx
# HjAcBgNVBAoTFU1pY3Jvc29mdCBDb3Jwb3JhdGlvbjEmMCQGA1UEAxMdTWljcm9z
# b2Z0IFRpbWUtU3RhbXAgUENBIDIwMTAwHhcNMjExMjAyMTkwNTEzWhcNMjMwMjI4
# MTkwNTEzWjCByjELMAkGA1UEBhMCVVMxEzARBgNVBAgTCldhc2hpbmd0b24xEDAO
# BgNVBAcTB1JlZG1vbmQxHjAcBgNVBAoTFU1pY3Jvc29mdCBDb3Jwb3JhdGlvbjEl
# MCMGA1UECxMcTWljcm9zb2Z0IEFtZXJpY2EgT3BlcmF0aW9uczEmMCQGA1UECxMd
# VGhhbGVzIFRTUyBFU046QUUyQy1FMzJCLTFBRkMxJTAjBgNVBAMTHE1pY3Jvc29m
# dCBUaW1lLVN0YW1wIFNlcnZpY2UwggIiMA0GCSqGSIb3DQEBAQUAA4ICDwAwggIK
# AoICAQDSH2wQC2+t/jzA6jL6LZMhDNJG0nv1cUqe+H4MGKyEgRZUwp1YsHl1ITGy
# i8K9rkPRKKKAi0lT8g0b1GIipkWc8qCtE3wibxoNR4mCyrvgEsXutnbxI1obx8cM
# fa2XgchG/XBGZcFtGd0UQvXkxUYvokfG1TyBMqnIZvQ2LtcmGj86laPRNuRodkEM
# 7VVUO2oMSHJbaTNj1b2kAC8sqlytH1zmfrQpTA3rZOyEmywT43DRfsNlXmkNKMiW
# 7BafNnHZLGHGacpimE4doDMur3yiH/qCCx2PO4pIqkA6WLGSN8yhYavcQZRFVtsl
# /x/IiuL0fxPGpQmRc84m41yauncveNh/5/14MqsZ7ugY1ix8fkOYgJBlLss8myPh
# aMA6qcEB/RWWqcCfhyARNjCcmBNGNXeMgKyZ/+e3bCOlXmWeDtVJDLmOtzEDBLmk
# g2/etp3T9hOX+LodYwdBkY2noCDEzPWVa834AmkJvR6ynEeBGj6ouWifpXxaobBd
# asb0+r/9eYr+T00yrLFn16rrTULnVzkW7lLyXWEousvzYnul3HPCQooQS4LY1HBK
# TyTSftGX56ZgOz7Rk+esvbcr+NjLvBBy7Xeomgkuw1F/Uru7lZ9AR+EQbpg2pvCH
# SarMQQHbf1GXPhlDTHwkeskRiz5jPjTr1Wz/f+9CZx5ovtTF0QIDAQABo4IBNjCC
# ATIwHQYDVR0OBBYEFNLfCNksLmWtIGEsiYuEKprRzXSyMB8GA1UdIwQYMBaAFJ+n
# FV0AXmJdg/Tl0mWnG1M1GelyMF8GA1UdHwRYMFYwVKBSoFCGTmh0dHA6Ly93d3cu
# bWljcm9zb2Z0LmNvbS9wa2lvcHMvY3JsL01pY3Jvc29mdCUyMFRpbWUtU3RhbXAl
# MjBQQ0ElMjAyMDEwKDEpLmNybDBsBggrBgEFBQcBAQRgMF4wXAYIKwYBBQUHMAKG
# UGh0dHA6Ly93d3cubWljcm9zb2Z0LmNvbS9wa2lvcHMvY2VydHMvTWljcm9zb2Z0
# JTIwVGltZS1TdGFtcCUyMFBDQSUyMDIwMTAoMSkuY3J0MAwGA1UdEwEB/wQCMAAw
# EwYDVR0lBAwwCgYIKwYBBQUHAwgwDQYJKoZIhvcNAQELBQADggIBAK9gCxC4IVbY
# KVQBHP5ztJc/kfgSubcL5hTReVE1uwSVKp92Sfd/IIvFgGQcwVlAZc8DubOhTshl
# R2fSFfK6+sUzoMOuf9ItKF7m348+SpZ455iITDyTgEjqXhTmTTvBfyEHA6gxHGzV
# o578k2Qsc7qSuXmPr8ZkeuRNHNOxFRQmnUWmdTOLGJlbJq9zTH+KYbnJZ2tK5xwT
# 2d2irtBu7U/FruzCxSbnM00y6dpYZcMUCdLuzxHEnX8/epO1nQlrpUTpJ6gel2Pv
# +E+4oktdX8zz0Y0WfwdQOZVbn5gr/wPLvIoceKJJ366AA36lbc8Do5h6TSvJbVAr
# Nutbg/1JcCT5Tl9peMEmiK1b3z5kRFZffztUe9pNYnhijkGaQnRTbsBqXaCCLmPU
# 9i4PEHcOyh8z7t5tzjOAnQYXi7oNBbRXitz8XbPK2XasNB9QaU+01TKZRlVtYlsW
# rDriN7xCwCcx4bUnyiHGNiV5reIsDMbCKZ7h1sxLIQeg5tW/Mg3R30EnzjFV5cq8
# RPXvoaFj89LpFMlmJbk8+KFmHzwXcl5wS+GVy38VulA+36aEM4FADKqMjW10FCUE
# VVfznFZ3UlGdSS7GqyFeoXBzEqvwaIWxv0BXvLtNPfR+YxOzeCaeiMVC3cx0PlDc
# z+AF/VN2WHKI81dOAmE/qLJkd/EpmLZzMIIHcTCCBVmgAwIBAgITMwAAABXF52ue
# AptJmQAAAAAAFTANBgkqhkiG9w0BAQsFADCBiDELMAkGA1UEBhMCVVMxEzARBgNV
# BAgTCldhc2hpbmd0b24xEDAOBgNVBAcTB1JlZG1vbmQxHjAcBgNVBAoTFU1pY3Jv
# c29mdCBDb3Jwb3JhdGlvbjEyMDAGA1UEAxMpTWljcm9zb2Z0IFJvb3QgQ2VydGlm
# aWNhdGUgQXV0aG9yaXR5IDIwMTAwHhcNMjEwOTMwMTgyMjI1WhcNMzAwOTMwMTgz
# MjI1WjB8MQswCQYDVQQGEwJVUzETMBEGA1UECBMKV2FzaGluZ3RvbjEQMA4GA1UE
# BxMHUmVkbW9uZDEeMBwGA1UEChMVTWljcm9zb2Z0IENvcnBvcmF0aW9uMSYwJAYD
# VQQDEx1NaWNyb3NvZnQgVGltZS1TdGFtcCBQQ0EgMjAxMDCCAiIwDQYJKoZIhvcN
# AQEBBQADggIPADCCAgoCggIBAOThpkzntHIhC3miy9ckeb0O1YLT/e6cBwfSqWxO
# dcjKNVf2AX9sSuDivbk+F2Az/1xPx2b3lVNxWuJ+Slr+uDZnhUYjDLWNE893MsAQ
# GOhgfWpSg0S3po5GawcU88V29YZQ3MFEyHFcUTE3oAo4bo3t1w/YJlN8OWECesSq
# /XJprx2rrPY2vjUmZNqYO7oaezOtgFt+jBAcnVL+tuhiJdxqD89d9P6OU8/W7IVW
# Te/dvI2k45GPsjksUZzpcGkNyjYtcI4xyDUoveO0hyTD4MmPfrVUj9z6BVWYbWg7
# mka97aSueik3rMvrg0XnRm7KMtXAhjBcTyziYrLNueKNiOSWrAFKu75xqRdbZ2De
# +JKRHh09/SDPc31BmkZ1zcRfNN0Sidb9pSB9fvzZnkXftnIv231fgLrbqn427DZM
# 9ituqBJR6L8FA6PRc6ZNN3SUHDSCD/AQ8rdHGO2n6Jl8P0zbr17C89XYcz1DTsEz
# OUyOArxCaC4Q6oRRRuLRvWoYWmEBc8pnol7XKHYC4jMYctenIPDC+hIK12NvDMk2
# ZItboKaDIV1fMHSRlJTYuVD5C4lh8zYGNRiER9vcG9H9stQcxWv2XFJRXRLbJbqv
# UAV6bMURHXLvjflSxIUXk8A8FdsaN8cIFRg/eKtFtvUeh17aj54WcmnGrnu3tz5q
# 4i6tAgMBAAGjggHdMIIB2TASBgkrBgEEAYI3FQEEBQIDAQABMCMGCSsGAQQBgjcV
# AgQWBBQqp1L+ZMSavoKRPEY1Kc8Q/y8E7jAdBgNVHQ4EFgQUn6cVXQBeYl2D9OXS
# ZacbUzUZ6XIwXAYDVR0gBFUwUzBRBgwrBgEEAYI3TIN9AQEwQTA/BggrBgEFBQcC
# ARYzaHR0cDovL3d3dy5taWNyb3NvZnQuY29tL3BraW9wcy9Eb2NzL1JlcG9zaXRv
# cnkuaHRtMBMGA1UdJQQMMAoGCCsGAQUFBwMIMBkGCSsGAQQBgjcUAgQMHgoAUwB1
# AGIAQwBBMAsGA1UdDwQEAwIBhjAPBgNVHRMBAf8EBTADAQH/MB8GA1UdIwQYMBaA
# FNX2VsuP6KJcYmjRPZSQW9fOmhjEMFYGA1UdHwRPME0wS6BJoEeGRWh0dHA6Ly9j
# cmwubWljcm9zb2Z0LmNvbS9wa2kvY3JsL3Byb2R1Y3RzL01pY1Jvb0NlckF1dF8y
# MDEwLTA2LTIzLmNybDBaBggrBgEFBQcBAQROMEwwSgYIKwYBBQUHMAKGPmh0dHA6
# Ly93d3cubWljcm9zb2Z0LmNvbS9wa2kvY2VydHMvTWljUm9vQ2VyQXV0XzIwMTAt
# MDYtMjMuY3J0MA0GCSqGSIb3DQEBCwUAA4ICAQCdVX38Kq3hLB9nATEkW+Geckv8
# qW/qXBS2Pk5HZHixBpOXPTEztTnXwnE2P9pkbHzQdTltuw8x5MKP+2zRoZQYIu7p
# Zmc6U03dmLq2HnjYNi6cqYJWAAOwBb6J6Gngugnue99qb74py27YP0h1AdkY3m2C
# DPVtI1TkeFN1JFe53Z/zjj3G82jfZfakVqr3lbYoVSfQJL1AoL8ZthISEV09J+BA
# ljis9/kpicO8F7BUhUKz/AyeixmJ5/ALaoHCgRlCGVJ1ijbCHcNhcy4sa3tuPywJ
# eBTpkbKpW99Jo3QMvOyRgNI95ko+ZjtPu4b6MhrZlvSP9pEB9s7GdP32THJvEKt1
# MMU0sHrYUP4KWN1APMdUbZ1jdEgssU5HLcEUBHG/ZPkkvnNtyo4JvbMBV0lUZNlz
# 138eW0QBjloZkWsNn6Qo3GcZKCS6OEuabvshVGtqRRFHqfG3rsjoiV5PndLQTHa1
# V1QJsWkBRH58oWFsc/4Ku+xBZj1p/cvBQUl+fpO+y/g75LcVv7TOPqUxUYS8vwLB
# gqJ7Fx0ViY1w/ue10CgaiQuPNtq6TPmb/wrpNPgkNWcr4A245oyZ1uEi6vAnQj0l
# lOZ0dFtq0Z4+7X6gMTN9vMvpe784cETRkPHIqzqKOghif9lwY1NNje6CbaUFEMFx
# BmoQtB1VM1izoXBm8qGCAs4wggI3AgEBMIH4oYHQpIHNMIHKMQswCQYDVQQGEwJV
# UzETMBEGA1UECBMKV2FzaGluZ3RvbjEQMA4GA1UEBxMHUmVkbW9uZDEeMBwGA1UE
# ChMVTWljcm9zb2Z0IENvcnBvcmF0aW9uMSUwIwYDVQQLExxNaWNyb3NvZnQgQW1l
# cmljYSBPcGVyYXRpb25zMSYwJAYDVQQLEx1UaGFsZXMgVFNTIEVTTjpBRTJDLUUz
# MkItMUFGQzElMCMGA1UEAxMcTWljcm9zb2Z0IFRpbWUtU3RhbXAgU2VydmljZaIj
# CgEBMAcGBSsOAwIaAxUA0PommlVZaduKtDHghztBZDfmVv6ggYMwgYCkfjB8MQsw
# CQYDVQQGEwJVUzETMBEGA1UECBMKV2FzaGluZ3RvbjEQMA4GA1UEBxMHUmVkbW9u
# ZDEeMBwGA1UEChMVTWljcm9zb2Z0IENvcnBvcmF0aW9uMSYwJAYDVQQDEx1NaWNy
# b3NvZnQgVGltZS1TdGFtcCBQQ0EgMjAxMDANBgkqhkiG9w0BAQUFAAIFAOa5teIw
# IhgPMjAyMjA4MzExODI2NDJaGA8yMDIyMDkwMTE4MjY0MlowdzA9BgorBgEEAYRZ
# CgQBMS8wLTAKAgUA5rm14gIBADAKAgEAAgINHQIB/zAHAgEAAgISDjAKAgUA5rsH
# YgIBADA2BgorBgEEAYRZCgQCMSgwJjAMBgorBgEEAYRZCgMCoAowCAIBAAIDB6Eg
# oQowCAIBAAIDAYagMA0GCSqGSIb3DQEBBQUAA4GBAGuD9IeUe8nYAt0P4l3egsHO
# Dz6LpTCPwOtriYk8TlQdiHg34taT/w7SelW1E1C/8R74LI2l/G+eXCIPi1Koz6RU
# tUl6Xk+RKaH8APmGzeA8/wWnXDr42YMv/TtiipuzDla6Uypk02uJ/hupxXRHQ9nG
# y8xhNWvTDgxyORugBtmpMYIEDTCCBAkCAQEwgZMwfDELMAkGA1UEBhMCVVMxEzAR
# BgNVBAgTCldhc2hpbmd0b24xEDAOBgNVBAcTB1JlZG1vbmQxHjAcBgNVBAoTFU1p
# Y3Jvc29mdCBDb3Jwb3JhdGlvbjEmMCQGA1UEAxMdTWljcm9zb2Z0IFRpbWUtU3Rh
# bXAgUENBIDIwMTACEzMAAAGWSVti4S/d908AAQAAAZYwDQYJYIZIAWUDBAIBBQCg
# ggFKMBoGCSqGSIb3DQEJAzENBgsqhkiG9w0BCRABBDAvBgkqhkiG9w0BCQQxIgQg
# 29sZgH+NC/wSy81mIaz9J64mVzDclRb4fdIHvlhtJgcwgfoGCyqGSIb3DQEJEAIv
# MYHqMIHnMIHkMIG9BCB2BNYC+B0105J2Ry6CfnZ0JA8JflZQQ6sLpHI3LbK9kDCB
# mDCBgKR+MHwxCzAJBgNVBAYTAlVTMRMwEQYDVQQIEwpXYXNoaW5ndG9uMRAwDgYD
# VQQHEwdSZWRtb25kMR4wHAYDVQQKExVNaWNyb3NvZnQgQ29ycG9yYXRpb24xJjAk
# BgNVBAMTHU1pY3Jvc29mdCBUaW1lLVN0YW1wIFBDQSAyMDEwAhMzAAABlklbYuEv
# 3fdPAAEAAAGWMCIEICbxovfCCRL8Mn13t9jSolnfB30lAGhhsTPKqxHVrqEOMA0G
# CSqGSIb3DQEBCwUABIICAAV6xYZ3P2WwKCNv3oaZoVTadRTYgBHc7UAUnwVuHkO0
# y3FNcMdhtAhIYJC+DKJrNCdI8QDCs9Tox36AIi3PE+7ZMf765my56eAsrdW/iErk
# DcTIUKbg0+ZIsmv/xxckPozfSigOeq6+Ma3N8+yOhXdsdUKx0kIIiyF77L9mHXnI
# hQoquEbsKSeO5QDy8M2U6BI0z2LKm9jNaLQU9IpBLN2jUtvNBcKexjOGpv0xLxaS
# 9VgRSM6YUrCrEGuFBMdxB8V0iu8Pzuf/EdzpQrtlso5CIiqyqOCgOl7UU4Zjd2Ys
# GzPd+4YOmd2bFDt9Nq02yxdR67hwQa/w8ct8QKTlj1wZ46Z0LH2zQ28LNSbZ3Ij5
# 5qxN+8yEnfMA59B2it8vtOhnzga0iDDK43UapBd9+CES1GMY+xcvnnDdisrbv7ip
# XSYzudJrJxtkhhJ1q7DYIBfb5Imztssuu78UNG1vhcpkA/qQdxs5RYJauzNEYBIS
# G5lvbKBCBzCShbebjTnb41ESUQu9uucZUS51UhXCtEgO5tt9sOfeL3/JdWKIlzv/
# gYPr2FHVfmqxywwZ+xQ78kbD4D5tuDBChXx+8jncRYAYUl31uyIRdwhNLrBP4DmV
# mxraJlNRrOrx1vMWFyYuoBhNRxZpyukjIazStYu+WKqprOwxm95Ca1/w0y2hVsfE
# SIG # End signature block
